/**
 * 
 * AiCiA - Android IRC Client
 * 
 * Copyright (C)2010 GORRY.
 * 
 */

package net.gorry.aicia;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import android.content.Context;
import android.content.SharedPreferences;
import android.text.SpannableStringBuilder;
import android.text.util.Linkify;
import android.util.Log;

/**
 * 
 * IRCサーバリスト処理
 * 
 * @author GORRY
 *
 */
public class IRCServerList {
	private static final String TAG = "IRCServerList";
	private static final boolean V = false;
	private static final boolean D = false;

	private final ArrayList<IRCServer> mIrcServerList = new ArrayList<IRCServer>();
	private final HashMap<String, Integer> mIrcServerNameMap = new HashMap<String, Integer>();
	private String mCurrentServerName;
	private IRCServerListEventListener[] mListeners = new IRCServerListEventListener[0];
	private final Context me;
	private ArrayList<SpannableStringBuilder> mSpanOtherChannelLog;
	private final boolean mConnect;

	/**
	 * コンストラクタ
	 * @param context コンテキスト
	 * @param connect IRCサーバに接続する実体として使用するときはtrue
	 */
	public IRCServerList(final Context context, final boolean connect) {
		if (V) Log.v(TAG, "IRCServerList()");
		me = context;
		mConnect = connect;
		mSpanOtherChannelLog = new ArrayList<SpannableStringBuilder>();
		echoMessageToOther("sub log");
	}

	/**
	 * リスナへのイベント転送処理
	 */
	private class IRCServerListener implements IRCServerEventListener {
		public synchronized void receiveMessageToChannel(final String serverName, final String channel, final String nick, final String dateMsg, final SpannableStringBuilder ssb, final boolean alert) {
			doReceiveMessageToChannel(serverName, channel, nick, dateMsg, ssb, alert);
		}

		public void createServer(final String serverName, final String dateMsg) {
			for (int i=mListeners.length-1; i>=0; i--) {
				mListeners[i].createServer(serverName, dateMsg);
			}
		}

		public void removeServer(final String serverName, final String dateMsg) {
			for (int i=mListeners.length-1; i>=0; i--) {
				mListeners[i].removeServer(serverName, dateMsg);
			}
		}

		public void createChannel(final String serverName, final String channel, final String dateMsg) {
			for (int i=mListeners.length-1; i>=0; i--) {
				mListeners[i].createChannel(serverName, channel, dateMsg);
			}
		}

		public void removeChannel(final String serverName, final String channel, final String dateMsg) {
			for (int i=mListeners.length-1; i>=0; i--) {
				mListeners[i].removeChannel(serverName, channel, dateMsg);
			}
		}

		public void changeNick(final String serverName, final String oldNick, final String newNick, final String dateMsg) {
			for (int i=mListeners.length-1; i>=0; i--) {
				mListeners[i].changeNick(serverName, oldNick, newNick, dateMsg);
			}
		}

		public void changeTopic(final String serverName, final String channel, final String topic, final String dateMsg) {
			for (int i=mListeners.length-1; i>=0; i--) {
				mListeners[i].changeTopic(serverName, channel, topic, dateMsg);
			}
		}

		public void receiveConnect(final String serverName) {
			for (int i=mListeners.length-1; i>=0; i--) {
				mListeners[i].receiveConnect(serverName);
			}
		}

		public void receiveDisconnect(final String serverName) {
			for (int i=mListeners.length-1; i>=0; i--) {
				mListeners[i].receiveDisconnect(serverName);
			}
		}
	}

	/**
	 * リスナ登録処理
	 * @param l リスナ
	 */
	public synchronized void addEventListener(final IRCServerListEventListener l) {
		if (V) Log.v(TAG, "addEventListener()");
		if (l == null) {
			throw new IllegalArgumentException("Listener is null.");
		}
		final int len = mListeners.length;
		final IRCServerListEventListener[] oldListeners = mListeners;
		mListeners = new IRCServerListEventListener[len + 1];
		System.arraycopy(oldListeners, 0, mListeners, 0, len);
		mListeners[len] = l;
	}

	/**
	 * リスナ削除処理
	 * @param l リスナ
	 * @return true 正常に削除された
	 */
	public synchronized boolean removeEventListener(final IRCServerListEventListener l) {
		if (V) Log.v(TAG, "removeEventListener()");
		if (l == null) {
			return false;
		}
		int index = -1;
		for (int i=0; i<mListeners.length; i++) {
			if (mListeners[i].equals(l)) {
				index = i;
				break;
			}
		}
		if (index == -1) {
			return false;
		}
		mListeners[index] = null;
		final int len = mListeners.length - 1;
		final IRCServerListEventListener[] newListeners = new IRCServerListEventListener[len];
		for (int i=0, j=0; i<len; j++) {
			if (mListeners[j] != null) {
				newListeners[i++] = mListeners[j];
			}
		}
		mListeners = newListeners;
		return true;
	}

	/**
	 * IRCメッセージ処理
	 * @param serverName サーバ設定名
	 * @param channel チャンネル名
	 * @param nick Nick
	 * @param dateMsg 日付時刻
	 * @param ssb Spanメッセージ
	 * @param alert アラートする必要があるときtrue
	 */
	public synchronized void doReceiveMessageToChannel(final String serverName, final String channel, final String nick, final String dateMsg, final SpannableStringBuilder ssb, final boolean alert) {
		int serverId = -1;
		int channelId = -1;
		serverId = getServerId(serverName);
		if (serverId >= 0) {
			channelId = mIrcServerList.get(serverId).getChannelId(channel);
		}

		// otherChannel用のログを生成
		final SpannableStringBuilder ssbOther = IRCMsg.colorDateMsg(dateMsg, SystemConfig.subLogDateColor);
		if ((channel != null) && (channel.length() > 0) && (!channel.equals(IRCMsg.sOtherChannelName))) {
			String ids = "";
			if ((0 <= serverId) && (serverId < 10)) {
				final String idServer = "1234567890";
				ids = idServer.substring(serverId, serverId+1);
				if ((1 <= channelId) && (channelId <= 26)) {  // 0はシステムチャンネルなので割り当てない
					final String idChannel = "abcdefghijklmnopqrstuvwxyz";
					ids = ids + idChannel.substring(channelId-1, channelId);
				}
				ids = ids + ":";
			}
			final SpannableStringBuilder ssbChannel = new SpannableStringBuilder("<" + ids + IRCMsg.channelName(channel) + "> ");
			ssbOther.append(ssbChannel);
		}
		if ((nick != null) && (nick.length() > 0)) {
			ssbOther.append(IRCMsg.colorNick(nick));
		}
		ssbOther.append(ssb);
		ssbOther.append("\n");

		boolean toOther = true;
		if ((serverName != null) && (channel != null) && (mCurrentServerName != null)) {
			final String currentChannel = getCurrentChannel(mCurrentServerName);
			if (currentChannel != null) {
				if (serverName.equalsIgnoreCase(mCurrentServerName) &&
						channel.equalsIgnoreCase(currentChannel)) {
					toOther = false;
				}
			}
		}
		if (toOther) {
			// spanOtherChannelLogへ収集
			final int size = mSpanOtherChannelLog.size();
			if (size >= SystemConfig.spanOtherChannelLogLines*2) {
				final ArrayList<SpannableStringBuilder> ssba = new ArrayList<SpannableStringBuilder>(
						mSpanOtherChannelLog.subList(size-SystemConfig.spanOtherChannelLogLines, size));
				mSpanOtherChannelLog = ssba;
			}
			mSpanOtherChannelLog.add(ssbOther);
		}

		for (int i=mListeners.length-1; i>=0; i--) {
			mListeners[i].receiveMessageToChannel(serverName, channel, nick, dateMsg, ssb, ssbOther, alert);
		}
	}

	/**
	 * サーバ名リストを得る
	 * @return サーバ名リスト
	 */
	public String[] getServerList() {
		final int numServers = mIrcServerList.size();
		final String[] serverList = new String[numServers];
		for (int i=0; i<numServers; i++) {
			serverList[i] = mIrcServerList.get(i).getServerName();
		}
		return serverList;
	}

	/**
	 * サーバ設定名からサーバIDを得る
	 * @param serverName サーバ設定名
	 * @return サーバID
	 */
	public int getServerId(final String serverName) {
		if (V) Log.v(TAG, "getServerId()");
		if (serverName == null) return -1;
		final Object obj = mIrcServerNameMap.get(serverName.toLowerCase());
		if (obj == null) {
			return -1;
		}
		return ((Integer)obj).intValue();
	}

	/**
	 * サーバIDからサーバ設定名を得る
	 * @param serverId サーバID
	 * @return サーバ設定名
	 */
	public String getServerName(final int serverId) {
		if (V) Log.v(TAG, "getServerName()");
		final int num = mIrcServerList.size();
		if ((0 <= serverId) && (serverId < num)) {
			return new String(mIrcServerList.get(serverId).getServerName());
		}
		return null;
	}

	/**
	 * サーバIDからサーバインスタンスを得る
	 * @param serverId サーバID
	 * @return サーバインスタンス
	 */
	public IRCServer getServer(final int serverId) {
		if (V) Log.v(TAG, "getServer()");
		final int num = mIrcServerList.size();
		if ((0 <= serverId) && (serverId < num)) {
			return mIrcServerList.get(serverId);
		}
		return null;
	}

	/**
	 * サーバ設定名からサーバインスタンスを得る
	 * @param serverName サーバ設定名
	 * @return サーバインスタンス
	 */
	public IRCServer getServer(final String serverName) {
		if (V) Log.v(TAG, "getServer()");
		final int serverId = getServerId(serverName);
		if (0 <= serverId) {
			return mIrcServerList.get(serverId);
		}
		return null;
	}

	/**
	 * 全サーバ設定の再読み込み
	 */
	public void reloadList() {
		if (V) Log.v(TAG, "reloadList()");
		mIrcServerList.clear();
		mIrcServerNameMap.clear();
		final SharedPreferences pref = me.getSharedPreferences("ircserverlist", 0);
		final int num = pref.getInt("numServers", 0);
		for (int i=0; i<num; i++) {
			final IRCServer ircServer = new IRCServer(me, mConnect);
			ircServer.addEventListener(new IRCServerListener());
			ircServer.loadConfig(i);
			mIrcServerList.add(ircServer);
			mIrcServerNameMap.put(ircServer.getServerName().toLowerCase(), i);
		}
	}

	/**
	 * ID指定でサーバ設定の再読み込み
	 * @param serverId サーバID
	 */
	public void reloadServerConfig(final int serverId) {
		if (V) Log.v(TAG, "reloadServerConfig()");
		final int num = mIrcServerList.size();
		if ((0 <= serverId) && (serverId < num)) {
			mIrcServerNameMap.remove(mIrcServerList.get(serverId).getServerName().toLowerCase());
			mIrcServerList.get(serverId).loadConfig(serverId);
			mIrcServerNameMap.put(mIrcServerList.get(serverId).getServerName().toLowerCase(), serverId);
			return;
		}
		if (D) throw new IllegalArgumentException("serverId [" + serverId + "] is not found");
	}

	/**
	 * ID指定でサーバの削除
	 * @param serverId サーバID
	 */
	public void removeServer(final int serverId) {
		if (V) Log.v(TAG, "removeServer()");
		final int num = mIrcServerList.size();
		if ((0 <= serverId) && (serverId < num)) {
			final String serverName = mIrcServerList.get(serverId).getServerName();
			// リストを詰める
			int j = 0;
			for (int i=0; i<num; i++) {
				if (i != serverId) {
					mIrcServerList.get(i).saveConfig(j);
					j++;
				}
			}
			final SharedPreferences pref = me.getSharedPreferences("ircserverlist", 0);
			final SharedPreferences.Editor editor = pref.edit();
			editor.putInt("numServers", num-1);
			editor.commit();

			// 末尾の削除
			IRCServerConfig config = new IRCServerConfig(me);
			config.deleteConfig(num);
			config = null;

			mIrcServerList.get(serverId).disconnect();
			mIrcServerList.remove(serverId);
			mIrcServerNameMap.remove(serverName.toLowerCase());

			// カレントサーバの調整
			if ( getServerId(mCurrentServerName) < 0 ) {
				final int newNum = mIrcServerList.size();
				final int newId = Math.min(serverId, newNum-1);
				if (newId < 0) {
					mCurrentServerName = null;
				} else {
					mCurrentServerName = mIrcServerList.get(newId).getServerName();
				}
			}
			return;
		}
		if (D) throw new IllegalArgumentException("serverId [" + serverId + "] not found");
	}

	/**
	 * サーバの追加
	 * @param config サーバ設定
	 */
	public void addServer(final IRCServerConfig config) {
		if (V) Log.v(TAG, "addServer()");
		if (config.mServerName == null) return;
		if (getServerId(config.mServerName) >= 0) {
			if (D) throw new IllegalArgumentException("serverName [" + config.mServerName + "] is already exist");
			return;
		}
		int num = mIrcServerList.size();
		config.saveConfig(num);
		final IRCServer ircServer = new IRCServer(me, mConnect);
		ircServer.addEventListener(new IRCServerListener());
		ircServer.loadConfig(num);
		mIrcServerList.add(ircServer);
		mIrcServerNameMap.put(ircServer.getServerName().toLowerCase(), num);
		num++;
		final SharedPreferences pref = me.getSharedPreferences("ircserverlist", 0);
		final SharedPreferences.Editor editor = pref.edit();
		editor.putInt("numServers", num);
		editor.commit();
	}

	/**
	 * IRCサーバへの接続発行
	 * @param serverName サーバ設定名
	 */
	public void connectServer(final String serverName) {
		if (V) Log.v(TAG, "connectServer()");
		final int serverId = getServerId(serverName);
		if (serverId < 0) {
			if (D) throw new IllegalArgumentException("serverName [" + serverName + "] is not exist");
			return;
		}
		final IRCServer ircServer = mIrcServerList.get(serverId);
		ircServer.connect();
	}

	/**
	 * IRCサーバへの切断発行
	 * @param serverName サーバ設定名
	 */
	public void disconnectServer(final String serverName) {
		if (V) Log.v(TAG, "disconnectServer()");
		final int serverId = getServerId(serverName);
		if (serverId < 0) {
			if (D) throw new IllegalArgumentException("serverName [" + serverName + "] is not exist");
			return;
		}
		final IRCServer ircServer = mIrcServerList.get(serverId);
		ircServer.disconnect();
	}

	/**
	 * IRCサーバへのクローズ発行
	 * @param serverName サーバ設定名
	 */
	public void closeServer(final String serverName) {
		if (V) Log.v(TAG, "closeServer()");
		final int serverId = getServerId(serverName);
		if (serverId < 0) {
			if (D) throw new IllegalArgumentException("serverName [" + serverName + "] is not exist");
			return;
		}
		final IRCServer ircServer = mIrcServerList.get(serverId);
		ircServer.close();
	}

	/**
	 * 自動接続登録されているIRCサーバへの接続発行
	 */
	public void connectAuto() {
		if (V) Log.v(TAG, "connectAuto()");
		final int num = mIrcServerList.size();
		for (int i=0; i<num; i++) {
			mIrcServerList.get(i).connectAuto();
		}
	}

	/**
	 * 全IRCサーバへの切断発行
	 */
	public void disconnectAll() {
		if (V) Log.v(TAG, "disconnectAll()");
		final int num = mIrcServerList.size();
		for (int i=0; i<num; i++) {
			mIrcServerList.get(i).disconnect();
		}
	}

	/**
	 * 全IRCサーバへのクローズ発行
	 */
	public void closeAll() {
		if (V) Log.v(TAG, "closeAll()");
		final int num = mIrcServerList.size();
		for (int i=0; i<num; i++) {
			mIrcServerList.get(i).close();
		}
	}

	/**
	 * 接続が切れているかどうか確認
	 * @param serverName サーバ設定名
	 * @return 接続が切れている（＆再接続が必要）ならtrue
	 */
	public boolean isDisconnected(final String serverName) {
		if (V) Log.v(TAG, "isDisconnected()");
		final int serverId = getServerId(serverName);
		if (serverId < 0) {
			if (D) throw new IllegalArgumentException("serverName [" + serverName + "] is not exist");
			return false;
		}
		final IRCServer ircServer = mIrcServerList.get(serverId);
		return ircServer.isDisconnected();
	}

	/**
	 * 現在登録サーバ数の取得
	 * @return サーバ数
	 */
	public int getNumServers() {
		if (V) Log.v(TAG, "getNumServers()");
		return mIrcServerList.size();
	}

	/**
	 * カレントサーバを得る
	 * @return サーバ設定名
	 */
	public String getCurrentServerName() {
		if (V) Log.v(TAG, "getCurrentServerName()");
		return mCurrentServerName;
	}

	/**
	 * カレントサーバの設定
	 * @param serverName サーバ設定名
	 */
	public void setCurrentServerName(final String serverName) {
		if (V) Log.v(TAG, "setCurrentServerNo()");
		final int serverId = getServerId(serverName);
		if (serverId < 0) {
			if (D) throw new IllegalArgumentException("serverName [" + serverName + "] is not exist");
			return;
		}
		mCurrentServerName = new String(serverName);
	}

	/**
	 * チャンネルへコマンドライン送信
	 * @param serverName サーバ設定名
	 * @param channel チャンネル名
	 * @param message コマンドライン
	 */
	public void sendCommandLine(final String serverName, final String channel, final String message) {
		if (V) Log.v(TAG, "sendCommandLine()");
		final int serverId = getServerId(serverName);
		if (serverId < 0) {
			if (D) throw new IllegalArgumentException("serverName [" + serverName + "] is not exist");
			return;
		}
		final IRCServer ircServer = mIrcServerList.get(serverId);
		ircServer.sendCommandLine(channel, message);
	}

	/**
	 * チャンネルへメッセージ送信
	 * @param serverName サーバ設定名
	 * @param channel チャンネル名
	 * @param message メッセージ
	 */
	public void sendMessageToChannel(final String serverName, final String channel, final String message) {
		if (V) Log.v(TAG, "sendMessageToChannel()");
		final int serverId = getServerId(serverName);
		if (serverId < 0) {
			if (D) throw new IllegalArgumentException("serverName [" + serverName + "] is not exist");
			return;
		}
		final IRCServer ircServer = mIrcServerList.get(serverId);
		ircServer.sendChannel(channel, message);
	}

	/**
	 * チャンネルへメッセージエコー
	 * @param serverName サーバ設定名
	 * @param channel チャンネル名
	 * @param message メッセージ
	 */
	public void echoMessageToChannel(final String serverName, final String channel, final String message) {
		if (V) Log.v(TAG, "echoMessageToChannel()");
		final int serverId = getServerId(serverName);
		if (serverId < 0) {
			if (serverName.equalsIgnoreCase(IRCMsg.sOtherChannelServerName)) {
				final SpannableStringBuilder ssb = new SpannableStringBuilder(message + "\n");
				Linkify.addLinks(ssb, Linkify.WEB_URLS);
				if (isTIGMode(serverName)) {
					final Linkify.MatchFilter hashMatchFilter = new Linkify.MatchFilter() {
						public boolean acceptMatch(final CharSequence s, final int start, final int end) {
							if (start == 0) return true;
							final char c = s.charAt(start-1);
							if (IRCMsg.isUrlChar(c)) return false;
							return true;
						}
					};
					final Linkify.TransformFilter hashTransformFilter = new Linkify.TransformFilter() {
						public final String transformUrl(final Matcher match, final String url) {
							return match.group(1);
						}
					};
					final Pattern pattern = Pattern.compile("#([^ <]+)");
					final String scheme = (SystemConfig.twitterSiteIsMobile ?  "http://m.twitter.com/searches?q=%23" : "http://twitter.com/search?q=%23");
					Linkify.addLinks(ssb, pattern, scheme, hashMatchFilter, hashTransformFilter);
				}
				final String dateMsg = IRCMsg.getDateMsg();
				doReceiveMessageToChannel(
						IRCMsg.sOtherChannelServerName,
						IRCMsg.sOtherChannelName,
						null,
						dateMsg,
						ssb,
						false
				);
			} else {
				if (D) throw new IllegalArgumentException("serverName [" + serverName + "] is not exist");
			}
			return;
		}
		final IRCServer ircServer = mIrcServerList.get(serverId);
		ircServer.echoChannel(channel, message, true);
	}

	/**
	 * otherチャンネルへメッセージエコー
	 * @param message メッセージ
	 */
	public void echoMessageToOther(final String message) {
		echoMessageToChannel(IRCMsg.sOtherChannelServerName, IRCMsg.sOtherChannelName, message);
	}

	/**
	 * 指定サーバ・チャンネルのSpanログ取得
	 * @param serverName サーバ設定名
	 * @param channel チャンネル名
	 * @return Spanログ
	 */
	public SpannableStringBuilder getSpanChannelLog(final String serverName, final String channel) {
		if (V) Log.v(TAG, "getChannelLog()");
		final SpannableStringBuilder msg = new SpannableStringBuilder();
		if (serverName.equals(IRCMsg.sOtherChannelName)){
			for (int i=0; i<mSpanOtherChannelLog.size(); i++) {
				msg.append(mSpanOtherChannelLog.get(i));
			}
			return msg;
		}
		final int serverId = getServerId(serverName);
		if (serverId < 0) {
			if (D) throw new IllegalArgumentException("serverName [" + serverName + "] is not exist");
			return null;
		}
		final IRCServer ircServer = mIrcServerList.get(serverId);
		return ircServer.getSpanChannelLog(channel);
	}

	/**
	 * 指定サーバのトピック取得
	 * @param serverName サーバ設定名
	 * @param channel チャンネル
	 * @return トピック
	 */
	public String getTopic(final String serverName, final String channel) {
		if (V) Log.v(TAG, "getNick()");
		final int serverId = getServerId(serverName);
		if (serverId < 0) {
			if (D) throw new IllegalArgumentException("serverName [" + serverName + "] is not exist");
			return null;
		}
		final IRCServer ircServer = mIrcServerList.get(serverId);
		return ircServer.getTopic(channel);
	}

	/**
	 * 指定サーバのnick取得
	 * @param serverName サーバ設定名
	 * @return nick
	 */
	public String getNick(final String serverName) {
		if (V) Log.v(TAG, "getNick()");
		final int serverId = getServerId(serverName);
		if (serverId < 0) {
			if (D) throw new IllegalArgumentException("serverName [" + serverName + "] is not exist");
			return null;
		}
		final IRCServer ircServer = mIrcServerList.get(serverId);
		return ircServer.getNick();
	}

	/**
	 * 指定サーバのカレントチャンネル設定
	 * @param serverName サーバ設定名
	 * @param channel チャンネル名
	 */
	public void setCurrentChannel(final String serverName, final String channel) {
		if (V) Log.v(TAG, "setCurrentChannel()");
		final int serverId = getServerId(serverName);
		if (serverId < 0) {
			if (D) throw new IllegalArgumentException("serverName [" + serverName + "] is not exist");
			return;
		}
		final IRCServer ircServer = mIrcServerList.get(serverId);
		ircServer.setCurrentChannel(channel);
	}

	/**
	 * 指定サーバのカレントチャンネルを得る
	 * @param serverName サーバ設定名
	 * @return チャンネル名
	 */
	public String getCurrentChannel(final String serverName) {
		if (V) Log.v(TAG, "getCurrentChannel()");
		final int serverId = getServerId(serverName);
		if (serverId < 0) {
			if (D) throw new IllegalArgumentException("serverName [" + serverName + "] is not exist");
			return null;
		}
		final IRCServer ircServer = mIrcServerList.get(serverId);
		return ircServer.getCurrentChannel();
	}

	/**
	 * 指定サーバのチャンネル名リストを得る
	 * @param serverName サーバ設定名
	 * @return チャンネル名リスト
	 */
	public String[] getChannelList(final String serverName) {
		final int serverId = getServerId(serverName);
		if (serverId < 0) {
			if (D) throw new IllegalArgumentException("serverName [" + serverName + "] is not exist");
			return null;
		}
		final IRCServer ircServer = mIrcServerList.get(serverId);
		return ircServer.getChannelNameList();
	}

	/**
	 * 指定サーバのチャンネル更新リストを得る
	 * @param serverName サーバ設定名
	 * @return チャンネル更新リスト
	 */
	public Boolean[] getChannelUpdatedList(final String serverName) {
		final int serverId = getServerId(serverName);
		if (serverId < 0) {
			if (D) throw new IllegalArgumentException("serverName [" + serverName + "] is not exist");
			return null;
		}
		final IRCServer ircServer = mIrcServerList.get(serverId);
		return ircServer.getChannelUpdatedList();
	}

	/**
	 * 指定サーバのチャンネル更新フラグの取得
	 * @param serverName サーバ設定名
	 * @param channel チャンネル名
	 * @return チャンネル更新フラグがonならtrue
	 */
	public boolean getChannelUpdated(final String serverName, final String channel) {
		final int serverId = getServerId(serverName);
		if (serverId < 0) {
			if (D) throw new IllegalArgumentException("serverName [" + serverName + "] is not exist");
			return false;
		}
		final IRCServer ircServer = mIrcServerList.get(serverId);
		return ircServer.getChannelUpdated(channel);
	}

	/**
	 * チャンネル更新フラグのクリア
	 * @param serverName サーバ設定名
	 * @param channel チャンネル名
	 */
	public void clearChannelUpdated(final String serverName, final String channel) {
		final int serverId = getServerId(serverName);
		if (serverId < 0) {
			if (D) throw new IllegalArgumentException("serverName [" + serverName + "] is not exist");
			return;
		}
		final IRCServer ircServer = mIrcServerList.get(serverId);
		ircServer.clearChannelUpdated(channel);
	}

	/**
	 * 再接続時刻登録
	 * @param serverName サーバ設定名
	 */
	public void setNextReconnectTime(final String serverName) {
		if (V) Log.v(TAG, "setNextReconnectTime()");
		final int serverId = getServerId(serverName);
		if (serverId < 0) {
			if (D) throw new IllegalArgumentException("serverName [" + serverName + "] is not exist");
			return;
		}
		final IRCServer ircServer = mIrcServerList.get(serverId);
		ircServer.setNextReconnectTime();
	}

	//
	/**
	 * 再接続時刻の取得
	 * @param serverName サーバ設定名
	 * @return 再接続時刻
	 */
	public long getNextReconnectTime(final String serverName) {
		if (V) Log.v(TAG, "getNextReconnectTime()");
		final int serverId = getServerId(serverName);
		if (serverId < 0) {
			if (D) throw new IllegalArgumentException("serverName [" + serverName + "] is not exist");
			return 0;
		}
		final IRCServer ircServer = mIrcServerList.get(serverId);
		return ircServer.getNextReconnectTime();
	}

	/**
	 * KeepAlive送出時刻登録
	 * @param serverName サーバ設定名
	 */
	public void setNextKeepAliveTime(final String serverName) {
		if (V) Log.v(TAG, "setNextKeepAliveTime()");
		final int serverId = getServerId(serverName);
		if (serverId < 0) {
			if (D) throw new IllegalArgumentException("serverName [" + serverName + "] is not exist");
			return;
		}
		final IRCServer ircServer = mIrcServerList.get(serverId);
		ircServer.setNextKeepAliveTime();
	}

	//
	/**
	 * KeepAlive送出時刻の取得
	 * @param serverName サーバ設定名
	 * @return 送出時刻
	 */
	public long getNextKeepAliveTime(final String serverName) {
		if (V) Log.v(TAG, "getNextKeepAliveTime()");
		final int serverId = getServerId(serverName);
		if (serverId < 0) {
			if (D) throw new IllegalArgumentException("serverName [" + serverName + "] is not exist");
			return 0;
		}
		final IRCServer ircServer = mIrcServerList.get(serverId);
		return ircServer.getNextKeepAliveTime();
	}

	//
	/**
	 * 自動再接続フラグの取得
	 * @param serverName サーバ設定名
	 * @return 自動再接続フラグがonならtrue
	 */
	public boolean getNeedReconnect(final String serverName) {
		if (V) Log.v(TAG, "getNeedReconnect()");
		final int serverId = getServerId(serverName);
		if (serverId < 0) {
			if (D) throw new IllegalArgumentException("serverName [" + serverName + "] is not exist");
			return false;
		}
		final IRCServer ircServer = mIrcServerList.get(serverId);
		return ircServer.getNeedReconnect();
	}

	/**
	 * 指定サーバのチャンネルに入る
	 * @param serverName サーバ設定名
	 * @param channel チャンネル名
	 */
	public void joinChannel(final String serverName, final String channel) {
		if (V) Log.v(TAG, "joinChannel()");
		final int serverId = getServerId(serverName);
		if (serverId < 0) {
			if (D) throw new IllegalArgumentException("serverName [" + serverName + "] is not exist");
			return;
		}
		final IRCServer ircServer = mIrcServerList.get(serverId);
		ircServer.joinChannel(channel);
	}

	/**
	 * 指定サーバの複数チャンネルに順次入る
	 * @param serverName サーバ設定名
	 * @param channels チャンネル名（スペース区切り）
	 */
	public void joinChannels(final String serverName, final String channels) {
		if (V) Log.v(TAG, "joinChannels()");
		final int serverId = getServerId(serverName);
		if (serverId < 0) {
			if (D) throw new IllegalArgumentException("serverName [" + serverName + "] is not exist");
			return;
		}
		final IRCServer ircServer = mIrcServerList.get(serverId);
		ircServer.joinChannels(channels);
	}

	/**
	 * 指定サーバのチャンネルから出る
	 * @param serverName サーバ設定名
	 * @param channel チャンネル名
	 */
	public void partChannel(final String serverName, final String channel) {
		if (V) Log.v(TAG, "partChannel()");
		final int serverId = getServerId(serverName);
		if (serverId < 0) {
			if (D) throw new IllegalArgumentException("serverName [" + serverName + "] is not exist");
			return;
		}
		final IRCServer ircServer = mIrcServerList.get(serverId);
		ircServer.partChannel(channel);
	}

	/**
	 * ユーザーリストを得る
	 * @param serverName サーバ設定名
	 * @param channel チャンネル名
	 * @return ユーザーリスト
	 */
	public String[] getUserList(final String serverName, final String channel) {
		if (V) Log.v(TAG, "getUserList()");
		final int serverId = getServerId(serverName);
		if (serverId < 0) {
			if (D) throw new IllegalArgumentException("serverName [" + serverName + "] is not exist");
			return null;
		}
		final IRCServer ircServer = mIrcServerList.get(serverId);
		return ircServer.getUserList(channel);
	}

	/**
	 * 指定サーバがTIGモードかどうかを得る
	 * @param serverName サーバ設定名
	 * @return TIGモードならtrue
	 */
	public boolean isTIGMode(final String serverName) {
		if (V) Log.v(TAG, "isTIGMode()");
		final int serverId = getServerId(serverName);
		if (serverId < 0) {
			if (D) throw new IllegalArgumentException("serverName [" + serverName + "] is not exist");
			return false;
		}
		final IRCServer ircServer = mIrcServerList.get(serverId);
		return ircServer.isTIGMode();
	}



}
